/*
    Grapple - A fully featured network layer with a simple interface
    Copyright (C) 2006 Michael Simms

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Michael Simms
    michael@linuxgamepublishing.com
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifdef HAVE_IFADDRS_H
#include <ifaddrs.h>
#endif

#include "grapple_client.h"
#include "grapple_defines.h"
#include "grapple_nattrav_internal.h"
#include "grapple_nattrav_client_internal.h"
#include "tools.h"

#ifndef HOST_NAME_MAX
# define HOST_NAME_MAX 255
#endif

extern int gethostname(char *name, size_t len);

static char **get_interface_list()
{
  char ipaddr[HOST_NAME_MAX+1];
  static char **interfaces=NULL;
  static int found_interfaces=0;
  int interface_count=0;

  if (found_interfaces)
    return interfaces;

  found_interfaces=1;

#ifdef HAVE_IFADDRS_H
  {
    struct ifaddrs *interface_scan = NULL, *interface_list = NULL;
    
    if (getifaddrs(&interface_list) < 0)
      {
	//We cant get the list of interfaces
	return;
      }
    
    interface_scan=interface_list;
    
    while (interface_scan)
      {
	if (interface_scan->ifa_addr->sa_family == AF_INET)
	  {
	    if (getnameinfo (interface_scan->ifa_addr, 
			     sizeof(struct sockaddr_in),
			     ipaddr, HOST_NAME_MAX, 
			     NULL, 0, NI_NUMERICHOST) == 0)
	      {
		if (interfaces)
		  interfaces=(char **)realloc(interfaces,
					     (interface_count+2)*sizeof(char *));
		else
		  interfaces=(char **)malloc(2*sizeof(char *));
		
		interfaces[interface_count]=(char *)malloc(strlen(ipaddr)+1);
		strcpy(interfaces[interface_count],ipaddr);
		
		interfaces[++interface_count]=NULL;
	      }
	  }
	interface_scan=interface_scan->ifa_next;
      }
    
    freeifaddrs(interface_list);
  }
#else
  //We dont have ifaddrs, thats a problem. For now, just use the one 
  //value that we get from gethostname
  {
    char hostname[HOST_NAME_MAX+1];
    struct hostent *hp;
    struct in_addr inet_address;
    
    
    gethostname(hostname,HOST_NAME_MAX);
    
    hp=gethostbyname(hostname);
    
    if (!hp)
      return NULL;
    
    interface_count=0;
    
    while (hp->h_addr_list[interface_count])
      {
	if (interfaces)
	  interfaces=(char **)realloc(interfaces,
				     (interface_count+2)*sizeof(char *));
	else
	  interfaces=(char **)malloc(2*sizeof(char *));
	
	memcpy((char *)&inet_address,hp->h_addr,sizeof(struct in_addr));
	
	strcpy(ipaddr,inet_ntoa(inet_address));
	
	interfaces[interface_count]=(char *)malloc(strlen(ipaddr)+1);
	strcpy(interfaces[interface_count],ipaddr);

	interfaces[++interface_count]=NULL;
      }
  }
#endif
  
  return interfaces;
}

static void process_user_msg_address_response(internal_server_data *serverdata,
					      grapple_message *message)
{
  intchar val;
  int port;
  char *host;
  char **interfaces;
  int ifcount;

  if (message->USER_MSG.length < 9)
    return;

  memcpy(val.c,message->USER_MSG.data+4,4);

  port=ntohl(val.i);

  host=(char *)malloc(message->USER_MSG.length-7);
  memcpy(host,message->USER_MSG.data+8,message->USER_MSG.length-8);
  host[message->USER_MSG.length-8]=0;

  //Now we test out whether this is the same as the address we are
  //binding to

  serverdata->published_ip=host;
  serverdata->published_port=port;

  //If the port has changed, we MUST be in some kind of NAT
  if (port!=serverdata->port)
    {
      return;
    }

  if (serverdata->ip && !strcmp(host,serverdata->ip))
    {
      //Open internet at specified address
      serverdata->socket_nat_type=GRAPPLE_NAT_OPEN;

      return;
    }

  //The port is the same, lets scan through all possible IP addresses for this
  //box
  interfaces=get_interface_list();

  if (!interfaces)
    return;

  //We have some interfaces, look into them
  ifcount=0;
  while (interfaces[ifcount])
    {
      if (!strcmp(interfaces[ifcount],host))
	{
	  //Its a local interface
	  serverdata->socket_nat_type=GRAPPLE_NAT_OPEN;

	  return;
	}
      ifcount++;
    }
}

int grapple_nattrav_client_serverside_run(internal_server_data *serverdata)
{ 
  grapple_client client;
  intchar val;
  grapple_message *message;
  int done;
  int messagetype;

  if (serverdata->published_ip)
    {
      free(serverdata->published_ip);
      serverdata->published_ip=NULL;
    }
  serverdata->socket_nat_type=GRAPPLE_NAT_UNKNOWN;

  if (serverdata->nattrav_server_port==0)
    return 0;

  //Create a new socket to talk to the nattrav server
  client=grapple_client_init(serverdata->productname,
			     serverdata->productversion);
  grapple_client_address_set(client,serverdata->nattrav_server_hostname);
  grapple_client_port_set(client,serverdata->nattrav_server_port);
  grapple_client_sourceport_set(client,serverdata->port);
  grapple_client_protocol_set(client,GRAPPLE_PROTOCOL_UDP);
  grapple_client_notified_set(client,GRAPPLE_NOTIFY_STATE_OFF);
  grapple_client_start(client,GRAPPLE_WAIT);
  grapple_client_name_set(client,"Ntv");

  if (!grapple_client_connected(client))
    return 0;

  val.i=htonl(GRAPPLE_NATTRAV_ADDRESS_REQUEST);

  //Send a request for the address
  grapple_client_send(client,GRAPPLE_SERVER,GRAPPLE_RELIABLE,
		      val.c,4);

  //Wait while we are connected for the response to cone back
  done=0;
  while (!done && grapple_client_connected(client))
    {
      if (grapple_client_messages_waiting(client))
	{
	  message=grapple_client_message_pull(client);
	  if (message->type==GRAPPLE_MSG_USER_MSG)
	    {
	      if (message->USER_MSG.length>=4)
		{
		  memcpy(val.c,message->USER_MSG.data,4);

		  messagetype=ntohl(val.i);
  
		  if (messagetype==GRAPPLE_NATTRAV_ADDRESS_RESPONSE)
		    {
		      process_user_msg_address_response(serverdata,
							message);
		      done=1;
		    }
		}
	    }
	  grapple_message_dispose(message);
	}
    }

  if (serverdata->socket_nat_type==GRAPPLE_NAT_UNKNOWN)
    {
      //printf("We are behind some kind of NAT\n");
    }


  grapple_client_stop(client);
  grapple_client_destroy(client);

  return 0;
}
