/*
    Grapple - A fully featured network layer with a simple interface
    Copyright (C) 2006 Michael Simms

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Michael Simms
    michael@linuxgamepublishing.com
*/

#define _XOPEN_SOURCE 500

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>

#include "grapple.h"
#include "grapple_error.h"
#include "grapple_nattrav_internal.h"
#include "grapple_nattrav_server.h"
#include "grapple_nattrav_server_thread.h"
#include "grapple_structs.h"
#include "grapple_thread.h"
#include "tools.h"

/**************************************************************************
 ** The functions in this file are generally those that are accessible   **
 ** to the end user. Obvious exceptions are those that are static which  **
 ** are just internal utilities.                                         **
 ** Care should be taken to not change the parameters of outward facing  **
 ** functions unless absolutely required                                 **
 **************************************************************************/

//This is a static variable which keeps track of the list of all nattrav 
//servers run by this program. The servers are kept in a linked list. This 
//variable is global to this file only.
static internal_nattrav_server_data *grapple_nattrav_server_head=NULL;

//And this is the mutex to make this threadsafe
static grapple_thread_mutex *nattrav_mutex;

static int init_nattrav_mutex(void)
{
  static int done=0;

  if (done==1)
    return 1;
  done=1;

  nattrav_mutex=grapple_thread_mutex_init();

  return 1;
}

static internal_nattrav_server_data *nattrav_create(void)
{
  static int nextval=1;
  internal_nattrav_server_data *nattravdata;

  //Create the structure
  nattravdata=(internal_nattrav_server_data *)calloc(1,sizeof(internal_nattrav_server_data));

  //Assign it a default ID
  nattravdata->nattravnum=nextval++;

  //Create the mutexes we'll need
  nattravdata->inuse=grapple_thread_mutex_init();

  return nattravdata;
}

static int internal_nattrav_link(internal_nattrav_server_data *data)
{
  grapple_thread_mutex_lock(nattrav_mutex,GRAPPLE_LOCKTYPE_EXCLUSIVE);
  if (!grapple_nattrav_server_head)
    {
      grapple_nattrav_server_head=data;
      data->next=data;
      data->prev=data;
      grapple_thread_mutex_unlock(nattrav_mutex);
      return 1;
    }

  data->next=grapple_nattrav_server_head;
  data->prev=grapple_nattrav_server_head->prev;
  data->next->prev=data;
  data->prev->next=data;

  grapple_nattrav_server_head=data;

  grapple_thread_mutex_unlock(nattrav_mutex);
  
  return 1;
}

grapple_nattrav_server grapple_nattrav_server_init(const char *name,const char *version)
{
  internal_nattrav_server_data *nattravdata;
  grapple_nattrav_server returnval;

  init_nattrav_mutex();

  //Create the internal data
  nattravdata=nattrav_create();

  //Assign the user supplied values
  nattravdata->productname=(char *)malloc(strlen(name)+1);
  strcpy(nattravdata->productname,name);

  nattravdata->productversion=(char *)malloc(strlen(version)+1);
  strcpy(nattravdata->productversion,version);

  returnval=nattravdata->nattravnum;

  //Link it into the array of nattrav servers
  internal_nattrav_link(nattravdata);

  //Return the nattrav ID - the end user only gets an integer, called a
  //'grapple_nattrav_server'

  return returnval;
}

internal_nattrav_server_data *internal_nattrav_get(grapple_nattrav_server num)
{
  internal_nattrav_server_data *scan;
  int finished=0,found;
  
  while (!finished)
    {
      //By default if passed 0, then the oldest nattrav is returned
      if (!num)
        {
          grapple_thread_mutex_lock(nattrav_mutex,GRAPPLE_LOCKTYPE_SHARED);

          if (!grapple_nattrav_server_head)
            {
              grapple_thread_mutex_unlock(nattrav_mutex);
              
              return NULL;
            }

          if (grapple_thread_mutex_trylock(grapple_nattrav_server_head->inuse,
					   GRAPPLE_LOCKTYPE_EXCLUSIVE)==0)
            {
              grapple_thread_mutex_unlock(nattrav_mutex);
              return grapple_nattrav_server_head;
            }
          grapple_thread_mutex_unlock(nattrav_mutex);
        }
      else
        {
          grapple_thread_mutex_lock(nattrav_mutex,
				    GRAPPLE_LOCKTYPE_SHARED);

          //Loop through the nattravs
          scan=grapple_nattrav_server_head;

          found=0;

          while (scan && !found)
            {
              if (scan->nattravnum==num)
                {
                  if (grapple_thread_mutex_trylock(scan->inuse,
						   GRAPPLE_LOCKTYPE_EXCLUSIVE)==0)
                    {
                      //Match and return it
                      grapple_thread_mutex_unlock(nattrav_mutex);
                      return scan;
                    }
                  //It is in use, we cant use it yet

                  //Mark it as found though so we dont exit
                  found=1;
                }
      
              scan=scan->next;
              if (scan==grapple_nattrav_server_head)
                scan=NULL;
            }
          grapple_thread_mutex_unlock(nattrav_mutex);

          if (!found)
            //It isnt here, return NULL
            return NULL;
            
          //It is here but in use, sleep a very small amount
          microsleep(1000);
        }
    }

  //We never get here
  return NULL;
}

static void internal_nattrav_release(internal_nattrav_server_data *target)
{
  //We dont need to mutex this, we definitely HAVE it, and we are just
  //releasing it, and it wont be referenced again - it cant be deleted like
  //this

  grapple_thread_mutex_unlock(target->inuse);
}


int grapple_nattrav_server_ip_set(grapple_nattrav_server nattrav,const char *ip)
{
  internal_nattrav_server_data *nattravdata;

  nattravdata=internal_nattrav_get(nattrav);

  if (!nattravdata)
    {
      return GRAPPLE_FAILED;
    }

  if (nattravdata->server)
    {
      grapple_nattrav_server_error_set(nattravdata,GRAPPLE_ERROR_SERVER_CONNECTED);
      internal_nattrav_release(nattravdata);
      return GRAPPLE_FAILED;
    }

  //Free the old data if set
  if (nattravdata->ip)
    free (nattravdata->ip);

  //Set the new value
  nattravdata->ip=(char *)malloc(strlen(ip)+1);
  strcpy(nattravdata->ip,ip);

  internal_nattrav_release(nattravdata);

  return GRAPPLE_OK;
}

int grapple_nattrav_server_ip2_set(grapple_nattrav_server nattrav,const char *ip)
{
  internal_nattrav_server_data *nattravdata;

  nattravdata=internal_nattrav_get(nattrav);

  if (!nattravdata)
    {
      return GRAPPLE_FAILED;
    }

  if (nattravdata->server)
    {
      grapple_nattrav_server_error_set(nattravdata,GRAPPLE_ERROR_SERVER_CONNECTED);
      internal_nattrav_release(nattravdata);
      return GRAPPLE_FAILED;
    }

  //Free the old data if set
  if (nattravdata->ip2)
    free (nattravdata->ip2);

  //Set the new value
  nattravdata->ip2=(char *)malloc(strlen(ip)+1);
  strcpy(nattravdata->ip2,ip);

  internal_nattrav_release(nattravdata);

  return GRAPPLE_OK;
}

int grapple_nattrav_server_port_set(grapple_nattrav_server nattrav,int port)
{
  internal_nattrav_server_data *nattravdata;

  //Get the nattrav data
  nattravdata=internal_nattrav_get(nattrav);

  if (!nattravdata)
    {
      return GRAPPLE_FAILED;
    }

  if (nattravdata->server)
    {
      grapple_nattrav_server_error_set(nattravdata,GRAPPLE_ERROR_SERVER_CONNECTED);
      internal_nattrav_release(nattravdata);
      return GRAPPLE_FAILED;
    }

  //Set the port
  nattravdata->port=port;

  internal_nattrav_release(nattravdata);

  return GRAPPLE_OK;
}

int grapple_nattrav_server_port2_set(grapple_nattrav_server nattrav,int port)
{
  internal_nattrav_server_data *nattravdata;

  //Get the nattrav data
  nattravdata=internal_nattrav_get(nattrav);

  if (!nattravdata)
    {
      return GRAPPLE_FAILED;
    }

  if (nattravdata->server)
    {
      grapple_nattrav_server_error_set(nattravdata,GRAPPLE_ERROR_SERVER_CONNECTED);
      internal_nattrav_release(nattravdata);
      return GRAPPLE_FAILED;
    }

  //Set the port
  nattravdata->port2=port;

  internal_nattrav_release(nattravdata);

  return GRAPPLE_OK;
}

int grapple_nattrav_server_start(grapple_nattrav_server nattrav)
{
  internal_nattrav_server_data *nattravdata;

  nattravdata=internal_nattrav_get(nattrav);

  //Check the nattravs minimum defaults are set
  if (!nattravdata)
    {
      return GRAPPLE_FAILED;
    }

  if (nattravdata->server)
    {
      grapple_nattrav_server_error_set(nattravdata,GRAPPLE_ERROR_SERVER_CONNECTED);
      internal_nattrav_release(nattravdata);
      return GRAPPLE_FAILED;
    }

  if (!nattravdata->port)
    {
      grapple_nattrav_server_error_set(nattravdata,GRAPPLE_ERROR_PORT_NOT_SET);
      internal_nattrav_release(nattravdata);
      return GRAPPLE_FAILED;
    }

  /*
  printf("Setting nattrav server up on %s %d\n",nattravdata->ip,
	 nattravdata->port);
  */
  nattravdata->server=grapple_server_init(nattravdata->productname,
					  nattravdata->productversion);

  if (nattravdata->ip)
    grapple_server_ip_set(nattravdata->server,nattravdata->ip);
  grapple_server_port_set(nattravdata->server,nattravdata->port);
  grapple_server_protocol_set(nattravdata->server,GRAPPLE_PROTOCOL_UDP);
  grapple_server_session_set(nattravdata->server,"Ntv");
  grapple_server_start(nattravdata->server);

  if (nattravdata->port2)
    {
      /*
      printf("Setting nattrav server 2 up on %s %d\n",nattravdata->ip2,
	     nattravdata->port2);
      */
      nattravdata->server2=grapple_server_init(nattravdata->productname,
					       nattravdata->productversion);
      if (nattravdata->ip2)
	grapple_server_ip_set(nattravdata->server2,nattravdata->ip2);
      grapple_server_port_set(nattravdata->server2,nattravdata->port2);
      grapple_server_protocol_set(nattravdata->server2,GRAPPLE_PROTOCOL_UDP);
      grapple_server_session_set(nattravdata->server2,"Ntv");
      grapple_server_start(nattravdata->server2);
    }

  //Start the server thread that will handle all the communication
  grapple_nattrav_server_thread_start(nattravdata);

  internal_nattrav_release(nattravdata);

  return GRAPPLE_OK;
}

void grapple_nattrav_server_error_set(internal_nattrav_server_data *nattrav,
			       grapple_error error)
{
  nattrav->last_error=error;
}
