/*
    Grapple - A fully featured network layer with a simple interface
    Copyright (C) 2006 Michael Simms

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Michael Simms
    michael@linuxgamepublishing.com
*/

#define _XOPEN_SOURCE 500
#include <string.h>
#include <stdlib.h>
#include <signal.h>
#include <stdio.h>
#include <errno.h>

#include "grapple_nattrav_server.h"
#include "grapple_nattrav_server_thread.h"
#include "grapple_nattrav_internal.h"
#include "grapple_structs.h"
#include "grapple_defines.h"
#include "tools.h"

//This is a static variable which keeps track of the list of all servers
//run by this program. The servers are kept in a linked list. This variable
//is global to this file only.
static nattrav_server_connection *nattrav_server_connection_head=NULL;

//And this is the mutex to make this threadsafe
static grapple_thread_mutex *nattrav_server_connection_mutex=NULL;

//Wrapper function for creating a new connection. Coded this way to allow
//the later implimentation of queues or stacks or something, if that turns
//out to be a timesaver
static nattrav_server_connection *nattrav_server_connection_aquire(void)
{
  nattrav_server_connection *newconn;

  newconn=(nattrav_server_connection *)calloc(1,sizeof(nattrav_server_connection));

  newconn->inuse=grapple_thread_mutex_init();

  newconn->connect_time=time(NULL);

  return newconn;
}

//Dispose of a connection. This wrapper frees all memory associated with a
//connection and closes any open sockets
static void nattrav_server_connection_dispose(nattrav_server_connection *connection)
{
  grapple_thread_mutex_destroy(connection->inuse);

  //Free the data
  free(connection);

  return;
}

//Find a connection somewhere in a list, by its ID number
static nattrav_server_connection *nattrav_server_connection_from_id(int id)
{
  nattrav_server_connection *scan;
  int finished=0,found;
  
  while (!finished)
    {
      //By default if passed 0, then the oldest server is returned
      if (!id)
	{
	  grapple_thread_mutex_lock(nattrav_server_connection_mutex,
				    GRAPPLE_LOCKTYPE_SHARED);

	  if (!nattrav_server_connection_head)
	    {
	      grapple_thread_mutex_unlock(nattrav_server_connection_mutex);
	      
	      return NULL;
	    }

	  if (grapple_thread_mutex_trylock(nattrav_server_connection_head->inuse,
				    GRAPPLE_LOCKTYPE_EXCLUSIVE)==0)
	    {
	      grapple_thread_mutex_unlock(nattrav_server_connection_mutex);
	      return nattrav_server_connection_head;
	    }
	  grapple_thread_mutex_unlock(nattrav_server_connection_mutex);
	}
      else
	{
	  grapple_thread_mutex_lock(nattrav_server_connection_mutex,
				    GRAPPLE_LOCKTYPE_SHARED);

	  //Loop through the servers
	  scan=nattrav_server_connection_head;

	  found=0;

	  while (scan && !found)
	    {
	      if (scan->userid==id)
		{
		  if (grapple_thread_mutex_trylock(scan->inuse,
						   GRAPPLE_LOCKTYPE_EXCLUSIVE)==0)
		    {
		      //Match and return it
		      grapple_thread_mutex_unlock(nattrav_server_connection_mutex);
		      return scan;
		    }
		  //It is in use, we cant use it yet

		  //Mark it as found though so we dont exit
		  found=1;
		}
      
	      scan=scan->next;
	      if (scan==nattrav_server_connection_head)
		scan=NULL;
	    }
	  grapple_thread_mutex_unlock(nattrav_server_connection_mutex);

	  if (!found)
	    //It isnt here, return NULL
	    return NULL;
	    
	  //It is here but in use, sleep a very small amount
	  microsleep(1000);
	}
    }

  //We never get here
  return NULL;
}

static void nattrav_server_connection_release(nattrav_server_connection *target)
{
  //We dont need to mutex this, we definitely HAVE it, and we are just
  //releasing it, and it wont be referenced again - it cant be deleted like
  //this

  grapple_thread_mutex_unlock(target->inuse);
}

static int init_nattrav_server_connection_mutex(void)
{
  if (!nattrav_server_connection_mutex)
    nattrav_server_connection_mutex=grapple_thread_mutex_init();

  return 1;
}

//Link a server to the list
static int nattrav_server_connection_link(nattrav_server_connection *data)
{
  grapple_thread_mutex_lock(nattrav_server_connection_mutex,
			    GRAPPLE_LOCKTYPE_EXCLUSIVE);
  if (!nattrav_server_connection_head)
    {
      nattrav_server_connection_head=data;
      data->next=data;
      data->prev=data;
      grapple_thread_mutex_unlock(nattrav_server_connection_mutex);
      return 1;
    }

  data->next=nattrav_server_connection_head;
  data->prev=nattrav_server_connection_head->prev;
  data->next->prev=data;
  data->prev->next=data;

  nattrav_server_connection_head=data;

  grapple_thread_mutex_unlock(nattrav_server_connection_mutex);
  
  return 1;
}

//Remove a server from the linked list
static int nattrav_server_connection_unlink(nattrav_server_connection *data)
{
  grapple_thread_mutex_lock(nattrav_server_connection_mutex,
			    GRAPPLE_LOCKTYPE_EXCLUSIVE);
  if (data->next==data)
    {
      nattrav_server_connection_head=NULL;
      grapple_thread_mutex_unlock(nattrav_server_connection_mutex);
      return 1;
    }

  data->next->prev=data->prev;
  data->prev->next=data->next;

  if (data==nattrav_server_connection_head)
    nattrav_server_connection_head=data->next;

  grapple_thread_mutex_unlock(nattrav_server_connection_mutex);

  //This is outside the list now so can be done unmutexed
  data->next=NULL;
  data->prev=NULL;

  return 1;
}



static int process_grapple_msg_user_disconnect(grapple_server server,
					       grapple_message *message)
{
  nattrav_server_connection *oldconn;

  oldconn=nattrav_server_connection_from_id(message->USER_DISCONNECTED.id);

  if (oldconn)
    {
      nattrav_server_connection_unlink(oldconn);
      nattrav_server_connection_release(oldconn);
      nattrav_server_connection_dispose(oldconn);
    }

  return 0;
}

static void process_user_msg_address_request(grapple_server server,
					     grapple_message *message)
{
  char *outbuf;
  char *host;
  int length;
  intchar val;

  host=grapple_server_client_address_get(server,message->USER_MSG.id);
  length=strlen(host);

  outbuf=(char *)malloc(length+9);

  val.i=htonl(GRAPPLE_NATTRAV_ADDRESS_RESPONSE);
  memcpy(outbuf,val.c,4);

  val.i=htonl(grapple_server_client_port_get(server,message->USER_MSG.id));
  memcpy(outbuf+4,val.c,4);
  strcpy(outbuf+8,host);
  
  grapple_server_send(server,message->USER_MSG.id,GRAPPLE_RELIABLE,
		      outbuf,length+8);

  free(outbuf);

  return;
}

static int process_grapple_msg_user_msg(grapple_server server,
					grapple_message *message)
{
  intchar val;
  grapple_nattrav_messagetype_internal messagetype;
  if (message->USER_MSG.length<4)
    return 0;
  
  memcpy(val.c,message->USER_MSG.data,4);

  messagetype=ntohl(val.i);
  
  switch (messagetype)
    {
    case GRAPPLE_NATTRAV_ADDRESS_REQUEST:
      process_user_msg_address_request(server,message);
      break;


    case GRAPPLE_NATTRAV_ADDRESS_RESPONSE:
      //Not used sending to server
      break;
    }

  return 0;
}


static int process_grapple_msg_new_user(grapple_server server,
					grapple_message *message)
{
  nattrav_server_connection *newconn;

  newconn=nattrav_server_connection_aquire();
  newconn->userid=message->NEW_USER.id;
  newconn->server=server;
  nattrav_server_connection_link(newconn);

  return 0;
}

//This is the function that processes a new packet.
static int grapple_nattrav_server_process_message(grapple_server server,
						  grapple_message *message)
{
  switch (message->type)
    {
    case GRAPPLE_MSG_NEW_USER:
      process_grapple_msg_new_user(server,message);
      break;
    case GRAPPLE_MSG_USER_DISCONNECTED:
      process_grapple_msg_user_disconnect(server,message);
      break;
    case GRAPPLE_MSG_USER_MSG:
      process_grapple_msg_user_msg(server,message);
      break;
    default:
      break;
    }
  return 1;
}

static void 
grapple_nattrav_server_thread_process(internal_nattrav_server_data *nattrav)
{
  grapple_message *message;

  while (grapple_server_messages_waiting(nattrav->server))
    {
      message=grapple_server_message_pull(nattrav->server);
      grapple_nattrav_server_process_message(nattrav->server,message);
      grapple_message_dispose(message);
    }
  if (nattrav->server2)
    {
      while (grapple_server_messages_waiting(nattrav->server2))
	{
	  message=grapple_server_message_pull(nattrav->server2);
	  grapple_nattrav_server_process_message(nattrav->server2,message);
	  grapple_message_dispose(message);
	}
    }
}


static void *grapple_nattrav_server_thread_main(void *voiddata)
{
  internal_nattrav_server_data *data;
  int finished=0;
  struct sigaction sa;

  memset(&sa,0,sizeof(struct sigaction));

  sa.sa_handler = SIG_IGN;
  sigaction(SIGPIPE, &sa, 0);

  //The nattrav we have started
  data=(internal_nattrav_server_data *)voiddata;

  //Continue while we are not finished
  while (!finished)
    {
      //Process the thread data (users etc)
      grapple_nattrav_server_thread_process(data);

      if (data->threaddestroy)
        {
          //We have been told to end the thread
          finished=1;

	}
    }

  //We're done, the thread ends when this function ends
  data->thread=0;
  data->threaddestroy=0;

  return NULL;
}

int grapple_nattrav_server_thread_start(internal_nattrav_server_data *data)
{
  int createval;

  init_nattrav_server_connection_mutex();

  data->threaddestroy=0;

  createval=-1;

  //Create the thread
  while(createval!=0)
    {
      createval=pthread_create(&data->thread,NULL,
                               grapple_nattrav_server_thread_main,
			       (void *)data);
      if (createval!=0)
        {
          if (errno!=EAGAIN)
            {
              //Problem creating the thread that isnt a case of 'it will work
              //later, dont create it
              return -1;
            }
        }
    }

  pthread_detach(data->thread);

  return 1;
}
